#define _KERNEL
#define _ASM_INLINES

/*
 * We use the preprocessor to rename the atomic_* inline functions to
 * __atomic_* for the ones we care about, and __unused_atomic_* for those
 * that we don't care about.
 *
 * Then, we #undef all the #defines.  This allows us to create plain ol'
 * functions with atomic_* names.  Inside these, we call the __atomic_*
 * inlines.  Finally, to conserve .text, we define the
 * char/short/int/long/ptr versions to be weak symbols pointing to one of
 * the specific size functions.
 */

#if defined(_LP64)
#define WRAP64(...)	WRAPXX(__VA_ARGS__)
#else
#define WRAP64(...)
#endif

#define atomic_inc_8		__atomic_inc_8
#define atomic_inc_16		__atomic_inc_16
#define atomic_inc_32		__atomic_inc_32
#define atomic_inc_64		__atomic_inc_64
#define atomic_inc_uchar	__unused_atomic_inc_uchar
#define atomic_inc_ushort	__unused_atomic_inc_ushort
#define atomic_inc_uint		__unused_atomic_inc_uint
#define atomic_inc_ulong	__unused_atomic_inc_ulong

#define atomic_dec_8		__atomic_dec_8
#define atomic_dec_16		__atomic_dec_16
#define atomic_dec_32		__atomic_dec_32
#define atomic_dec_64		__atomic_dec_64
#define atomic_dec_uchar	__unused_atomic_dec_uchar
#define atomic_dec_ushort	__unused_atomic_dec_ushort
#define atomic_dec_uint		__unused_atomic_dec_uint
#define atomic_dec_ulong	__unused_atomic_dec_ulong

#define atomic_add_8		__atomic_add_8
#define atomic_add_16		__atomic_add_16
#define atomic_add_32		__atomic_add_32
#define atomic_add_64		__atomic_add_64
#define atomic_add_char		__unused_atomic_add_char
#define atomic_add_short	__unused_atomic_add_short
#define atomic_add_int		__unused_atomic_add_int
#define atomic_add_long		__unused_atomic_add_long
#define atomic_add_ptr		__unused_atomic_add_ptr

#define atomic_or_8		__atomic_or_8
#define atomic_or_16		__atomic_or_16
#define atomic_or_32		__atomic_or_32
#define atomic_or_64		__atomic_or_64
#define atomic_or_uchar		__unused_atomic_or_uchar
#define atomic_or_ushort	__unused_atomic_or_ushort
#define atomic_or_uint		__unused_atomic_or_uint
#define atomic_or_ulong		__unused_atomic_or_ulong

#define atomic_and_8		__atomic_and_8
#define atomic_and_16		__atomic_and_16
#define atomic_and_32		__atomic_and_32
#define atomic_and_64		__atomic_and_64
#define atomic_and_uchar	__unused_atomic_and_uchar
#define atomic_and_ushort	__unused_atomic_and_ushort
#define atomic_and_uint		__unused_atomic_and_uint
#define atomic_and_ulong	__unused_atomic_and_ulong

#define atomic_swap_8		__atomic_swap_8
#define atomic_swap_16		__atomic_swap_16
#define atomic_swap_32		__atomic_swap_32
#define atomic_swap_64		__atomic_swap_64
#define atomic_swap_uchar	__unused_atomic_swap_uchar
#define atomic_swap_ushort	__unused_atomic_swap_ushort
#define atomic_swap_uint	__unused_atomic_swap_uint
#define atomic_swap_ulong	__unused_atomic_swap_ulong

#define atomic_cas_8		__atomic_cas_8
#define atomic_cas_16		__atomic_cas_16
#define atomic_cas_32		__atomic_cas_32
#define atomic_cas_64		__atomic_cas_64
#define atomic_cas_uchar	__unused_atomic_cas_uchar
#define atomic_cas_ushort	__unused_atomic_cas_ushort
#define atomic_cas_uint		__unused_atomic_cas_uint
#define atomic_cas_ulong	__unused_atomic_cas_ulong
#define atomic_cas_ptr		__unused_atomic_cas_ptr

#include <sys/atomic.h>

#undef atomic_inc_8
#undef atomic_inc_16
#undef atomic_inc_32
#undef atomic_inc_64
#undef atomic_inc_uchar
#undef atomic_inc_ushort
#undef atomic_inc_uint
#undef atomic_inc_ulong

#undef atomic_dec_8
#undef atomic_dec_16
#undef atomic_dec_32
#undef atomic_dec_64
#undef atomic_dec_uchar
#undef atomic_dec_ushort
#undef atomic_dec_uint
#undef atomic_dec_ulong

#undef atomic_add_8
#undef atomic_add_16
#undef atomic_add_32
#undef atomic_add_64
#undef atomic_add_char
#undef atomic_add_short
#undef atomic_add_int
#undef atomic_add_long
#undef atomic_add_ptr

#undef atomic_or_8
#undef atomic_or_16
#undef atomic_or_32
#undef atomic_or_64
#undef atomic_or_uchar
#undef atomic_or_ushort
#undef atomic_or_uint
#undef atomic_or_ulong

#undef atomic_and_8
#undef atomic_and_16
#undef atomic_and_32
#undef atomic_and_64
#undef atomic_and_uchar
#undef atomic_and_ushort
#undef atomic_and_uint
#undef atomic_and_ulong

#undef atomic_swap_8
#undef atomic_swap_16
#undef atomic_swap_32
#undef atomic_swap_64
#undef atomic_swap_uchar
#undef atomic_swap_ushort
#undef atomic_swap_uint
#undef atomic_swap_ulong

#undef atomic_cas_8
#undef atomic_cas_16
#undef atomic_cas_32
#undef atomic_cas_64
#undef atomic_cas_uchar
#undef atomic_cas_ushort
#undef atomic_cas_uint
#undef atomic_cas_ulong
#undef atomic_cas_ptr

#define WRAPXX(fxn, type)			\
	void fxn(volatile type *a)	\
	{				\
		__##fxn(a);		\
	}

WRAPXX(atomic_inc_8,      uint8_t)
WRAPXX(atomic_inc_16,     uint16_t)
WRAPXX(atomic_inc_32,     uint32_t)
WRAP64(atomic_inc_64,     uint64_t)
#pragma weak atomic_inc_uchar = atomic_inc_8
#pragma weak atomic_inc_ushort = atomic_inc_16
#pragma weak atomic_inc_uint = atomic_inc_32
#if defined(_LP64)
#pragma weak atomic_inc_ulong = atomic_inc_64
#else
#pragma weak atomic_inc_ulong = atomic_inc_32
#endif

WRAPXX(atomic_dec_8,      uint8_t)
WRAPXX(atomic_dec_16,     uint16_t)
WRAPXX(atomic_dec_32,     uint32_t)
WRAP64(atomic_dec_64,     uint64_t)
#pragma weak atomic_dec_uchar = atomic_dec_8
#pragma weak atomic_dec_ushort = atomic_dec_16
#pragma weak atomic_dec_uint = atomic_dec_32
#if defined(_LP64)
#pragma weak atomic_dec_ulong = atomic_dec_64
#else
#pragma weak atomic_dec_ulong = atomic_dec_32
#endif

#undef WRAPXX

#define WRAPXX(fxn, type1, type2)		\
	void fxn(volatile type1 *a, type2 b)	\
	{				\
		__##fxn(a, b);		\
	}

WRAPXX(atomic_add_8,     uint8_t,  int8_t)
WRAPXX(atomic_add_16,    uint16_t, int16_t)
WRAPXX(atomic_add_32,    uint32_t, int32_t)
WRAP64(atomic_add_64,    uint64_t, int64_t)
#pragma weak atomic_add_uchar = atomic_add_8
#pragma weak atomic_add_ushort = atomic_add_16
#pragma weak atomic_add_uint = atomic_add_32
#if defined(_LP64)
#pragma weak atomic_add_ulong = atomic_add_64
#else
#pragma weak atomic_add_ulong = atomic_add_32
#endif
#pragma weak atomic_add_ptr = atomic_add_ulong

WRAPXX(atomic_or_8,      uint8_t,  uint8_t)
WRAPXX(atomic_or_16,     uint16_t, uint16_t)
WRAPXX(atomic_or_32,     uint32_t, uint32_t)
WRAP64(atomic_or_64,     uint64_t, uint64_t)
#pragma weak atomic_or_uchar = atomic_or_8
#pragma weak atomic_or_ushort = atomic_or_16
#pragma weak atomic_or_uint = atomic_or_32
#if defined(_LP64)
#pragma weak atomic_or_ulong = atomic_or_64
#else
#pragma weak atomic_or_ulong = atomic_or_32
#endif

WRAPXX(atomic_and_8,      uint8_t,  uint8_t)
WRAPXX(atomic_and_16,     uint16_t, uint16_t)
WRAPXX(atomic_and_32,     uint32_t, uint32_t)
WRAP64(atomic_and_64,     uint64_t, uint64_t)
#pragma weak atomic_and_uchar = atomic_and_8
#pragma weak atomic_and_ushort = atomic_and_16
#pragma weak atomic_and_uint = atomic_and_32
#if defined(_LP64)
#pragma weak atomic_and_ulong = atomic_and_64
#else
#pragma weak atomic_and_ulong = atomic_and_32
#endif

#undef WRAPXX

#define WRAPXX(fxn, type)				\
	type fxn(volatile type *a, type b, type c)	\
	{						\
		return __##fxn(a, b, c);		\
	}

WRAPXX(atomic_cas_8,      uint8_t)
WRAPXX(atomic_cas_16,     uint16_t)
WRAPXX(atomic_cas_32,     uint32_t)
WRAP64(atomic_cas_64,     uint64_t)
#pragma weak atomic_cas_uchar = atomic_cas_8
#pragma weak atomic_cas_ushort = atomic_cas_16
#pragma weak atomic_cas_uint = atomic_cas_32
#if defined(_LP64)
#pragma weak atomic_cas_ulong = atomic_cas_64
#else
#pragma weak atomic_cas_ulong = atomic_cas_32
#endif
#pragma weak atomic_cas_ptr = atomic_cas_ulong

#undef WRAPXX

#define WRAPXX(fxn, type)				\
	type fxn(volatile type *a, type b)		\
	{						\
		return __##fxn(a, b);			\
	}

WRAPXX(atomic_swap_8,      uint8_t)
WRAPXX(atomic_swap_16,     uint16_t)
WRAPXX(atomic_swap_32,     uint32_t)
WRAP64(atomic_swap_64,     uint64_t)
#pragma weak atomic_swap_uchar = atomic_swap_8
#pragma weak atomic_swap_ushort = atomic_swap_16
#pragma weak atomic_swap_uint = atomic_swap_32
#if defined(_LP64)
#pragma weak atomic_swap_ulong = atomic_swap_64
#else
#pragma weak atomic_swap_ulong = atomic_swap_32
#endif
